/*----------------------------------------------------------------------------

  UWP/WinRT Interoperation sample

  Copyright(C) Web Technology Corp, all rights reserved.
  Code by Yoshiaki "Koumei" Hashimoto

----------------------------------------------------------------------------*/

#define STRICT

#include <assert.h>
#include <thread>

#include <windows.h>

#include <wrl/client.h>
#include <wrl/implements.h>
#include <wrl/event.h>
#include <wrl/wrappers/corewrappers.h>

#include <Windows.System.h>
#include <Windows.Foundation.h>
#include <Windows.UI.Notifications.h>

#include <appmodel.h>

#include <VersionHelpers.h>

#pragma comment(lib, "runtimeobject.lib")

/*----------------------------------------------------------------------------
	ݓ삵ĂAvfXNgbvubW[h𒲂ׂ
	ipbP[WID̗LŔ肷o[Wj
----------------------------------------------------------------------------*/
bool api_IsBridgedAppMode_1(bool& bBridgedMode)
{
	// {AIsWindows10OrGreater() 邽߂ɂ̓}jtFXgł̐錾KvA
	// ubW[h̏ꍇ͏ɓ\Ȃ̂ŋCɂȂėǂ
	if(!IsWindows10OrGreater())
	{
		bBridgedMode = false;
		return true;
	}

	// Windows 7Ƃ̌݊̂߁AGetCurrentPackageFullName() API͓IɎ擾
	decltype(&GetCurrentPackageFullName) fnGetCurrentPackageFullName;
	fnGetCurrentPackageFullName = (decltype(fnGetCurrentPackageFullName))GetProcAddress(GetModuleHandle(L"kernel32"), "GetCurrentPackageFullName");
	if(!fnGetCurrentPackageFullName)
	{
		// IsWindows10OrGreater() ʂȂ̂ŁA͑zOG[ƍl
		return false;
	}

	// pbP[W̎擾s
	WCHAR szPackageName[1024];
	UINT32 cchPackageName = _countof(szPackageName);
	LONG lResult = fnGetCurrentPackageFullName(&cchPackageName, szPackageName);

	if(lResult == APPMODEL_ERROR_NO_PACKAGE)
	{
		// pbP[W݂ȂƂG[FubW[hł͂Ȃ
		bBridgedMode = false;
		return true;
	}
	else if(lResult == ERROR_SUCCESS || lResult == ERROR_INSUFFICIENT_BUFFER)
	{
		// pbP[W擾łFubW[hł
		bBridgedMode = true;
		return true;
	}

	// ȊÕG[R[h͏s
	return false;
}

/*----------------------------------------------------------------------------
	ݓ삵ĂAvfXNgbvubW[h𒲂ׂ
	i^CAPI̓ۂŔ肷o[Wj
----------------------------------------------------------------------------*/
bool api_IsBridgedAppMode_2(bool& bBridgedMode)
{
	// {AIsWindows10OrGreater() 邽߂ɂ̓}jtFXgł̐錾KvA
	// ubW[h̏ꍇ͏ɓ\Ȃ̂ŋCɂȂėǂ
	if(!IsWindows10OrGreater())
	{
		bBridgedMode = false;
		return true;
	}

	bool bRet = false;
	bBridgedMode = false;

	std::thread worker([&]()
	{
		// CXbhOleInitializeȂǂĂĂ悤A[J[Xbhŏ
		HRESULT hr = Windows::Foundation::Initialize(RO_INIT_MULTITHREADED);
		if(FAILED(hr))
		{
			assert(false);
			return;
		}

		using Microsoft::WRL::ComPtr;
		using namespace Microsoft::WRL::Wrappers;
		using namespace ABI::Windows::UI::Notifications;
		using namespace ABI::Windows::Foundation;

		{ // Uninitialize()̑OfactoryReleaseKvȂ̂ŃubN
			ComPtr<ITileUpdateManagerStatics> factory;
			hr = GetActivationFactory(HStringReference(RuntimeClass_Windows_UI_Notifications_TileUpdateManager).Get(), &factory);
			if(SUCCEEDED(hr))
			{
				ComPtr<ITileUpdater> tile;
				hr = factory->CreateTileUpdaterForApplication(&tile);
				if(SUCCEEDED(hr))
				{
					bBridgedMode = true;
				}
			}
		}

		Windows::Foundation::Uninitialize();

		bRet = true;
	});
	worker.join();

	return bRet;
}

/*----------------------------------------------------------------------------
	ILauncherStatics2::QueryUriSupportAsync()Ă
----------------------------------------------------------------------------*/
bool api_QueryUriSupport(LPCWSTR lpszUri, bool& bSupported)
{
	bool bRet = false;
	std::thread worker([&]()
	{
		// [J[Xbhŏ
		HRESULT hr = Windows::Foundation::Initialize(RO_INIT_MULTITHREADED);
		if(FAILED(hr))
		{
			assert(false);
			return;
		}

		using Microsoft::WRL::ComPtr;
		using namespace Microsoft::WRL::Wrappers;
		using namespace ABI::Windows::System;
		using namespace ABI::Windows::Foundation;

		ComPtr<ILauncherStatics2> launcherStatics;
		hr = GetActivationFactory(HStringReference(RuntimeClass_Windows_System_Launcher).Get(), &launcherStatics);
		if(SUCCEEDED(hr))
		{
			ComPtr<IUriRuntimeClassFactory> uriFactory;
			hr = GetActivationFactory(HStringReference(RuntimeClass_Windows_Foundation_Uri).Get(), &uriFactory);
			if(SUCCEEDED(hr))
			{
				HString uriHString;
				hr = uriHString.Set(lpszUri);
				if(SUCCEEDED(hr))
				{
					ComPtr<IUriRuntimeClass> uri;
					hr = uriFactory->CreateUri(uriHString.Get(), &uri);
					if(SUCCEEDED(hr))
					{
						Event completed(CreateEventEx(nullptr, nullptr, CREATE_EVENT_MANUAL_RESET, WRITE_OWNER | EVENT_ALL_ACCESS));
						if(completed.IsValid())
						{
							__FIAsyncOperation_1_Windows__CSystem__CLaunchQuerySupportStatus* operation = nullptr;
							hr = launcherStatics->QueryUriSupportAsync(uri.Get(), LaunchQuerySupportType::LaunchQuerySupportType_Uri, &operation);
							if(SUCCEEDED(hr))
							{
								hr = operation->put_Completed
								(
									Microsoft::WRL::Callback<IAsyncOperationCompletedHandler<LaunchQuerySupportStatus>>
									(
										[&](IAsyncOperation<LaunchQuerySupportStatus>* pHandler, AsyncStatus /*status*/)->HRESULT
										{
											// check status
											LaunchQuerySupportStatus ret = LaunchQuerySupportStatus::LaunchQuerySupportStatus_NotSupported;
											if(SUCCEEDED(pHandler->GetResults(&ret)))
											{
												bSupported = (LaunchQuerySupportStatus::LaunchQuerySupportStatus_Available == ret) ? true : false;
											}
											SetEvent(completed.Get());
											bRet = true;
											return S_OK;
										}
									).Get()
								);

								if(SUCCEEDED(hr))
								{
									WaitForSingleObject(completed.Get(), INFINITE);
								}
							}
						}
					}
				}
			}
		}

		Windows::Foundation::Uninitialize();
	});
	worker.join();

	return bRet;
}

/*----------------------------------------------------------------------------
	ILauncherStatics::LaunchUriAsync()Ă
----------------------------------------------------------------------------*/
bool api_LaunchUri(LPCWSTR lpszUri)
{
	bool bRet = false;
	std::thread worker([lpszUri, &bRet]()
	{
		// [J[Xbhŏ
		HRESULT hr = Windows::Foundation::Initialize(RO_INIT_MULTITHREADED);
		if(FAILED(hr))
		{
			assert(false);
			return;
		}

		using Microsoft::WRL::ComPtr;
		using namespace Microsoft::WRL::Wrappers;
		using namespace ABI::Windows::System;
		using namespace ABI::Windows::Foundation;

		ComPtr<ILauncherStatics> launcherStatics;
		hr = GetActivationFactory(HStringReference(RuntimeClass_Windows_System_Launcher).Get(), &launcherStatics);
		if(SUCCEEDED(hr))
		{
			ComPtr<IUriRuntimeClassFactory> uriFactory;
			hr = GetActivationFactory(HStringReference(RuntimeClass_Windows_Foundation_Uri).Get(), &uriFactory);
			if(SUCCEEDED(hr))
			{
				HString uriHString;
				hr = uriHString.Set(lpszUri);
				if(SUCCEEDED(hr))
				{
					ComPtr<IUriRuntimeClass> uri;
					hr = uriFactory->CreateUri(uriHString.Get(), &uri);
					if(SUCCEEDED(hr))
					{
						Event launchCompleted(CreateEventEx(nullptr, nullptr, CREATE_EVENT_MANUAL_RESET, WRITE_OWNER | EVENT_ALL_ACCESS));
						if(launchCompleted.IsValid())
						{
							__FIAsyncOperation_1_boolean* operation = nullptr;
							hr = launcherStatics->LaunchUriAsync(uri.Get(), &operation);
							if(SUCCEEDED(hr))
							{
								hr = operation->put_Completed
								(
									Microsoft::WRL::Callback<IAsyncOperationCompletedHandler<bool>>
									(
										[&launchCompleted, &bRet](IAsyncOperation<bool>* pHandler, AsyncStatus /*status*/)->HRESULT
										{
											// check status
											boolean ret = false;
											if(SUCCEEDED(pHandler->GetResults(&ret)))
											{
												bRet = ret ? true : false;
											}
											SetEvent(launchCompleted.Get());
											return S_OK;
										}
									).Get()
								);

								if(SUCCEEDED(hr))
								{
									WaitForSingleObject(launchCompleted.Get(), INFINITE);
								}
							}
						}
					}
				}
			}
		}

		Windows::Foundation::Uninitialize();
	});
	worker.join();

	return bRet;
}
